@@ -3,15 +3,9 @@ import pathlib
 import sys
 import numpy as np
 from typing import List
-
-# CUDA stuff
-# TODO: move from pvnrtc module to cuda entirely to reduce dependencies
-from pynvrtc.compiler import Program
 from cuda import cuda, nvrtc
-
 from .arg_info import ArgInfo, verify_args
 from .callable_func import CallableFunc
-from .gpu_headers import CUDA_HEADER_MAP
 from .hat_file import Function
 
 
@@ -43,38 +37,64 @@ def _find_cuda_incl_path() -> pathlib.Path:
 
     return cuda_path
 
-
 def compile_cuda_program(cuda_src_path: pathlib.Path, func_name):
     src = cuda_src_path.read_text()
 
-    prog = Program(src=src, name=func_name, headers=CUDA_HEADER_MAP.values(), include_names=CUDA_HEADER_MAP.keys())
-    ptx = prog.compile([
-        '-use_fast_math',
-        '-default-device',
-        '-std=c++11'
-    ])
+    opts = [
+        # https://docs.nvidia.com/cuda/nvrtc/index.html#group__options
+        b'--gpu-architecture=compute_86',
+        b'--ptxas-options=--warn-on-spills', # https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#options-for-passing-specific-phase-options-ptxas-options
+        b'-use_fast_math',
+        b'--include-path=/usr/local/cuda-11.6/include/',
+        b'-std=c++17',
+        b'-default-device',
+        #b'--restrict',
+        #b'--device-int128'
+    ]
+
+    # Create program
+    err, prog = nvrtc.nvrtcCreateProgram(str.encode(src), func_name.encode('utf-8'), 0, [], [])
+    ASSERT_DRV(err)
+
+    # Compile program
+    err = nvrtc.nvrtcCompileProgram(prog, len(opts), opts)
+    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
+        err, log_size = nvrtc.nvrtcGetProgramLogSize(prog)
+        ASSERT_DRV(err)
+
+        log = "0" * log_size
+        e_log = log.encode('utf-8')
+        err = nvrtc.nvrtcGetProgramLog(prog, e_log)
+        print(e_log.decode('utf-8'))
+
+    # Get PTX from compilation
+    err, ptxSize = nvrtc.nvrtcGetPTXSize(prog)
+    ASSERT_DRV(err)
+    ptx = b" " * ptxSize
+    err = nvrtc.nvrtcGetPTX(prog, ptx)
 
     return ptx
 
 
-def initialize_cuda():
+def initialize_cuda(gpu_id):
     # Initialize CUDA Driver API
     err, = cuda.cuInit(0)
     ASSERT_DRV(err)
 
-    # Retrieve handle for device 0
-    # TODO: add support for multiple CUDA devices?
-    err, cuDevice = cuda.cuDeviceGet(0)
+    err, cuDevice = cuda.cuDeviceGet(gpu_id)
     ASSERT_DRV(err)
 
     # Create context
+    # TODO: USE cuDevicePrimaryCtxRetain for faster intialization
     err, context = cuda.cuCtxCreate(0, cuDevice)
     ASSERT_DRV(err)
+    return context
 
 
 def get_func_from_ptx(ptx, func_name):
-    # Note: Incompatible --gpu-architecture would be detected here
-    err, ptx_mod = cuda.cuModuleLoadData(ptx.encode('utf-8'))
+    # Load PTX as module data and retrieve function
+    ptx = np.char.array(ptx)
+    err, ptx_mod = cuda.cuModuleLoadData(ptx)
     ASSERT_DRV(err)
     err, kernel = cuda.cuModuleGetFunction(ptx_mod, func_name.encode('utf-8'))
     ASSERT_DRV(err)
@@ -82,57 +102,52 @@ def get_func_from_ptx(ptx, func_name):
     return kernel
 
 
-def _cuda_transfer_mem(usage, func, source_args: List, dest_args: List, arg_infos: List[ArgInfo], stream=None):
+def _cuda_transfer_mem(usage, func, source_args: List, dest_args: List, arg_infos: List[ArgInfo]):
     for source_arg, dest_arg, arg_info in zip(source_args, dest_args, arg_infos):
         if usage in arg_info.usage.value:
-            if stream:
-                err, = func(dest_arg, source_arg, arg_info.total_byte_size, stream)
-            else:
-                err, = func(dest_arg, source_arg, arg_info.total_byte_size)
+            err, = func(dest_arg, source_arg, arg_info.total_byte_size)
             ASSERT_DRV(err)
 
 
-def transfer_mem_host_to_cuda(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo], stream=None):
+def transfer_mem_host_to_cuda(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
     _cuda_transfer_mem(
         usage='input',
-        func=cuda.cuMemCpyHtoDAsync if stream else cuda.cuMemcpyHtoD,
+        func=cuda.cuMemcpyHtoD,
         source_args=[a.ctypes.data for a in host_args],
         dest_args=device_args,
-        arg_infos=arg_infos,
-        stream=stream
+        arg_infos=arg_infos
     )
 
 
-def transfer_mem_cuda_to_host(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo], stream=None):
+def transfer_mem_cuda_to_host(device_args: List, host_args: List[np.array], arg_infos: List[ArgInfo]):
     _cuda_transfer_mem(
         usage='output',
-        func=cuda.cuMemcpyDtoHAsync if stream else cuda.cuMemcpyDtoH,
+        func=cuda.cuMemcpyDtoH,
         source_args=device_args,
         dest_args=[a.ctypes.data for a in host_args],
-        arg_infos=arg_infos,
-        stream=stream
+        arg_infos=arg_infos
     )
 
 
-def allocate_cuda_mem(arg_infos: List[ArgInfo], stream=None):
+def allocate_cuda_mem(arg_infos: List[ArgInfo]):
     device_mem = []
 
     for arg in arg_infos:
         size = arg.total_byte_size
-        err, mem = cuda.cuMemAllocAsync(size, stream) if stream else cuda.cuMemAlloc(size)
+        err, mem = cuda.cuMemAlloc(size)
         try:
             ASSERT_DRV(err)
         except:
-            free_cuda_mem(device_mem, stream)
+            free_cuda_mem(device_mem)
             raise
         device_mem.append(mem)
 
     return device_mem
 
 
-def free_cuda_mem(args, stream=None):
+def free_cuda_mem(args):
     for arg in args:
-        cuda.cuMemFreeAsync(arg, stream) if stream else cuda.cuMemFree(arg)
+        cuda.cuMemFree(arg)
 
 
 def device_args_to_ptr_list(device_args: List):
@@ -162,9 +177,10 @@ class CudaCallableFunc(CallableFunc):
         self.stop_event = None
         self.exec_time = 0.
         self.cuda_src_path = cuda_src_path
+        self.context = None
 
-    def init_runtime(self):
-        initialize_cuda()
+    def init_runtime(self, benchmark: bool, gpu_id: int):
+        self.context = initialize_cuda(gpu_id)
 
         ptx = _PTX_CACHE.get(self.cuda_src_path)
         if not ptx:
@@ -172,13 +188,16 @@ class CudaCallableFunc(CallableFunc):
 
         self.kernel = get_func_from_ptx(ptx, self.func_name)
 
-    def cleanup_runtime(self):
-        pass
+    def cleanup_runtime(self, benchmark: bool):
+        cuda.cuCtxDestroy(self.context)
 
-    def init_main(self, warmup_iters=0, args=[]):
+    def init_main(self, benchmark: bool, warmup_iters=0, args=[], gpu_id: int=0):
         verify_args(args, self.arg_infos, self.func_name)
         self.device_mem = allocate_cuda_mem(self.arg_infos)
-        transfer_mem_host_to_cuda(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
+
+        if not benchmark:
+            transfer_mem_host_to_cuda(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
+
         self.ptrs = device_args_to_ptr_list(self.device_mem)
 
         err, self.start_event = cuda.cuEventCreate(cuda.CUevent_flags.CU_EVENT_DEFAULT)
@@ -195,12 +214,11 @@ class CudaCallableFunc(CallableFunc):
                 self.ptrs.ctypes.data,    # kernel arguments
                 0,    # extra (ignore)
             )
-            ASSERT_DRV(err)
-        else:
-            err, = cuda.cuCtxSynchronize()
-            ASSERT_DRV(err)
 
-    def main(self, iters=1, batch_size=1, args=[]) -> float:
+            if not benchmark:
+                ASSERT_DRV(err)
+
+    def main(self, benchmark: bool, iters=1, batch_size=1, args=[]) -> float:
         batch_timings: List[float] = []
         for _ in range(batch_size):
             err, = cuda.cuEventRecord(self.start_event, 0)
@@ -215,7 +233,9 @@ class CudaCallableFunc(CallableFunc):
                     self.ptrs.ctypes.data,    # kernel arguments
                     0,    # extra (ignore)
                 )
-                ASSERT_DRV(err)
+
+                if not benchmark:
+                    ASSERT_DRV(err)
 
             err, = cuda.cuEventRecord(self.stop_event, 0)
             ASSERT_DRV(err)
@@ -226,18 +246,19 @@ class CudaCallableFunc(CallableFunc):
             batch_timings.append(batch_time)
             self.exec_time += batch_time
 
-            err, = cuda.cuCtxSynchronize()
-            ASSERT_DRV(err)
+            if not benchmark:
+                err, = cuda.cuCtxSynchronize()
+                ASSERT_DRV(err)
 
         self.exec_time /= (iters * batch_size)
         return batch_timings
 
-    def cleanup_main(self, args=[]):
+    def cleanup_main(self, benchmark: bool, args=[]):
         # If there's no device mem, that means allocation during initialization failed, which means nothing else needs to be cleaned up either
-        if self.device_mem:
+        if not benchmark and self.device_mem:
             transfer_mem_cuda_to_host(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
+        if self.device_mem:
             free_cuda_mem(self.device_mem)
-
         err, = cuda.cuCtxSynchronize()
         ASSERT_DRV(err)
 
