@@ -37,12 +37,19 @@ def get_func_from_rocm_program(rocm_program, func_name):
     kernel = hipModuleGetFunction(rocm_module, func_name)
     return kernel
 
+cached_mem=[]
 
-def allocate_rocm_mem(arg_infos: List[ArgInfo]):
+def allocate_rocm_mem(benchmark: bool, arg_infos: List[ArgInfo], gpu_id: int):
     device_mem = []
     for arg in arg_infos:
         try:
-            mem = hipMalloc(arg.total_byte_size)
+            memory_cache = cached_mem[gpu_id]
+            if benchmark and arg.total_byte_size in memory_cache:
+                mem = memory_cache[arg.total_byte_size]
+            else:
+                mem = hipMalloc(arg.total_byte_size)
+                if benchmark:
+                    memory_cache[arg.total_byte_size] = mem
         except:
             free_rocm_mem(device_mem)
             raise
@@ -96,8 +103,15 @@ class RocmCallableFunc(CallableFunc):
         self.exec_time = 0.
         self.rocm_src_path = rocm_src_path
 
-    def init_runtime(self):
-        initialize_rocm()
+    def init_runtime(self, benchmark: bool, gpu_id: int):
+        if not benchmark:
+            initialize_rocm()
+
+        hipSetDevice(gpu_id)
+
+        # Add a separate cache for each gpu since device memory is not shareable (duh!)
+        while len(cached_mem) <= gpu_id:
+            cached_mem.append({})
 
         rocm_program = _HSACO_CACHE.get(self.rocm_src_path)
         if not rocm_program:
@@ -105,13 +119,15 @@ class RocmCallableFunc(CallableFunc):
 
         self.kernel = get_func_from_rocm_program(rocm_program, self.func_name)
 
-    def cleanup_runtime(self):
+    def cleanup_runtime(self, benchmark: bool):
         pass
 
-    def init_main(self, warmup_iters=0, args=[]):
+    def init_main(self, benchmark: bool, warmup_iters=0, args=[], gpu_id: int=0):
         verify_args(args, self.arg_infos, self.func_name)
-        self.device_mem = allocate_rocm_mem(self.arg_infos)
-        transfer_mem_host_to_rocm(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
+        self.device_mem = allocate_rocm_mem(benchmark, self.arg_infos, gpu_id)
+
+        if not benchmark:
+            transfer_mem_host_to_rocm(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
 
         class DataStruct(ctypes.Structure):
             _fields_ = [(f"arg{i}", ctypes.c_void_p) for i in range(len(self.arg_infos))]
@@ -129,10 +145,8 @@ class RocmCallableFunc(CallableFunc):
                 0,    # stream
                 self.data,    # data
             )
-        else:
-            hipDeviceSynchronize()
 
-    def main(self, iters=1, batch_size=1, args=[]) -> float:
+    def main(self, benchmark: bool, iters=1, batch_size=1, args=[]) -> float:
         batch_timings: List[float] = []
         for _ in range(batch_size):
             hipEventRecord(self.start_event)
@@ -152,17 +166,17 @@ class RocmCallableFunc(CallableFunc):
             batch_timings.append(batch_time)
             self.exec_time += batch_time
 
-            hipDeviceSynchronize()
+            if not benchmark:
+                hipDeviceSynchronize()
 
         self.exec_time /= (iters * batch_size)
         return batch_timings
 
-    def cleanup_main(self, args=[]):
+    def cleanup_main(self, benchmark: bool, args=[]):
         # If there's no device mem, that means allocation during initialization failed, which means nothing else needs to be cleaned up either
-        if self.device_mem:
+        if not benchmark and self.device_mem:
             transfer_mem_rocm_to_host(device_args=self.device_mem, host_args=args, arg_infos=self.arg_infos)
             free_rocm_mem(self.device_mem)
-
         hipDeviceSynchronize()
 
         if self.start_event:
